PROGRAM mpi_mod_INITBCG_vI_2D

IMPLICIT NONE
INCLUDE 'mpif.h'
INCLUDE 'mkl.fi'

! "==============================================================================="
! "                   Parallel-Projection Block Conjugate Gradient                "
! "                             Version 1.0, 12/22/2017                           "
! "                University of Minnesota, Twin Cities, MN 55414, USA            "
! "                           contact: kalan019@umn.edu                           "
! "==============================================================================="

! Example run
!------------
! mpirun -np 8 ./ppbcg.exe 8000 1.0D-6 1.0D-6 1000 40 20 20 0.6D0 2.0D0 1 2 4

! = = = = = = = = = = = = = = = = = = =

! This software is part of the original source files used in the paper
! "A Scalable Dense Linear System Solver for Multiple Right-Hand Sides in Data Analytics"
! written by V. Kalantzis, C. Malossi, C. Bekas, A. Curioni, E. Gallopoulos, and Y. Saad.
!
! This software is free; you can redistribute it and/or modify it under 
! the terms of the GNU Lesser General Public License as published by the 
! Free Software Foundation; either version 2.1 of the License, or (at 
! your option) any later version.
!
! The present software is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of 
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser 
! General Public License for more details.
!
! You should have received a copy of the GNU Lesser General 
! Public License along with this library; if not, write to 
! the Free Software Foundation, Inc., 51 Franklin St, Fifth 
! Floor, Boston, MA 02110-1301, USA.

!----------------!
!Define variables!
!----------------!
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE     :: A
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE     :: b, Ab
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE     :: x
DOUBLE PRECISION                                  :: tol1, tol2, theta, kappa
CHARACTER(len=32)                                 :: BUFFER
INTEGER                                           :: N, nx, ny, len, IERR, I, J, CHUNK1, CHUNK2, indx, one
INTEGER                                           :: RANK, RANK1, SZ, blk, col_rank, row_rank, GINDEX1, GINDEX2
INTEGER                                           :: maxs, choice, row_comm, col_comm, p, q, INDEX1, INDEX2
INTEGER                                           :: rhs, seedb1, seedb2, maxit, cg_iters, bcg_iters
INTEGER                                           :: modinitbcg_iters, LIFO, tau
DOUBLE PRECISION                                  :: t1, t2, t10, t20
INTEGER                                           :: STATUSR(10), BNM
INTEGER                                           :: NP(2)
LOGICAL                                           :: PERIOD(2),ORDER,S_COL(2),S_ROW(2)
INTEGER                                           :: CART_GRP, MEOGRP, MEPGRP 
INTEGER(2)                                        :: COO
DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE       :: timings_cg, timings_bcg
DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE       :: timings_modinitbcg0, timings_mm
INTEGER                                           :: cg_matvecs, bcg_matvecs, modinitbcg_full_matvecs
DOUBLE PRECISION                                  :: MV_FLOPS, DOTAXPY_FLOPS

!---------------------------!
!Load command line arguments!
!---------------------------!

! Size of the covariance matrix
CALL GETARG(1, BUFFER)
READ (BUFFER, "(i)") N

! Accuracy requested for the seed block
CALL GETARG(2, BUFFER)
READ (BUFFER, "(f)") tol1

! Accuracy requested for each block, tol1 < = tol2
CALL GETARG(3, BUFFER)
READ (BUFFER, "(f)") tol2

! max number of blocks retained in pp-bcg
CALL GETARG(4, BUFFER)
READ (BUFFER, "(i)") maxs

! Number of right-hand sides
CALL GETARG(5, BUFFER)
READ (BUFFER, "(i)") rhs

! Size of the seed-batch
CALL GETARG(6, BUFFER)
READ (BUFFER, "(i)") seedb1

! Size of each following batch
CALL GETARG(7, BUFFER)
READ (BUFFER, "(i)") seedb2

! Diagonal weight for the covariance matrix
CALL GETARG(8, BUFFER)
READ (BUFFER, "(f)") theta

! Off-diagonal weight for the covariance matrix
CALL GETARG(9, BUFFER)
READ (BUFFER, "(f)") kappa

! Choice of solver
CALL GETARG(10, BUFFER)
READ (BUFFER, "(i)") choice

! Number of rows of the 2-D processors' grid
CALL GETARG(11, BUFFER)
READ (BUFFER, "(i)") nx

! Number of columns of the 2-D processors' grid
CALL GETARG(12, BUFFER)
READ (BUFFER, "(i)") ny

!-----------------------------------!
! Check validity of input arguments !
!-----------------------------------!
IF ( ( MOD(N,nx) /= 0D0 ) .OR. ( MOD(N,ny) /= 0D0 ) ) THEN
  WRITE(*,*) "PROGRAM IS ABORTED, SIZE N IS NOT DIVISABLE BY nx OR ny, OR BOTH" 
  STOP
ENDIF

IF ( ( N <= 0 ) .OR. ( rhs <= 0 ) .OR. ( seedb1 <= 0 ) .OR. ( seedb2 <= 0 ) .OR. ( tol1 <= 0 ) .OR. ( tol2 <= 0 ) ) THEN
  WRITE(*,*) "PROGRAM IS ABORTED, ONE OR MORE OF THE FOLLOWING VARIABLES WAS NEGATIVE: N, rhs, seedb1, seedb2, maxit, tol1, tol2" 
  STOP
ENDIF

IF ( ( nx < 1 ) .OR. ( ny < 1 ) ) THEN
  WRITE(*,*) "PROGRAM IS ABORTED. nx, OR ny, OR BOTH, WAS-WERE LESS THAN ONE" 
  STOP
ENDIF

IF ( rhs .NE. (seedb1+seedb2) ) THEN
  WRITE(*,*) "PROGRAM IS ABORTED. RHS .NE. seedb1+seedb2" 
  STOP
ENDIF

!--------------------------!
!Initialize MPI environment!
!--------------------------!
CALL MPI_INIT(IERR)

! Let each processor retrieve its rank, as well as
! the total number of processors in the 2-D grid 
CALL MPI_COMM_RANK(MPI_COMM_WORLD, RANK1, IERR)
CALL MPI_COMM_SIZE(MPI_COMM_WORLD, SZ, IERR)

IF (RANK1 == 0) THEN
 IF ( SZ /= (nx*ny) ) THEN
  WRITE(*,*) "PROGRAM IS ABORTED, TOTAL NUMBER OF MPI PROCESSES DOES NOT EQUAL THE PRODUCT nx*ny" 
  STOP
 ENDIF
ENDIF

! Determine local chunks
CHUNK1 = N / nx
CHUNK2 = N / ny

! Define maximum number of iterations
maxit = 1000D0

!-----------------!
!Print header line!
!-----------------!
IF (RANK1 .EQ. 0) THEN
   WRITE(*, *) " "
   WRITE(*, *) "==============================================================================="
   WRITE(*, *) "                   Parallel-Projection Block Conjugate Gradient                "
   WRITE(*, *) "                             Version 1.0, 12/22/2017                           "
   WRITE(*, *) "                University of Minnesota, Twin Cities, MN 55414, USA            "
   WRITE(*, *) "         Developers: Vassilis Kalantzis, Cristiano Malossi, Costas Bekas       "
   WRITE(*, *) "             Efstratios Gallopoulos, Alessandro Curioni, Yousef Saad           "
   WRITE(*, *) "==============================================================================="
   WRITE(*, *) " "
ENDIF


!---------------!
!Print arguments!
!---------------!
IF (RANK1 .EQ. 0) THEN
   write(*,*) "==============================================================================="
   write(*,*) "Number of processors - Size of Chunks: ", SZ, CHUNK1, CHUNK2
   write(*,*) " "
   write(*,*) "List of arguments:"
   write(*,*) "Dimension of A: ", N
   write(*,*) "# of right-hand sides: ", rhs
   write(*,*) "Size of the seed-block", seedb1
   write(*,*) "Tolerance for the block-seed system: ", tol1
   write(*,*) "Tolerance for the following blocks: ", tol2
   write(*,*) "# of maximum iterations: ", maxit
   write(*,*) "Values of theta, kappa: ", theta, "--", kappa
   write(*,*) "2D grid of Processors:", nx, ny
   write(*,*) "==============================================================================="
ENDIF

! Parameters for creating the 2D grid of processors
NP(1)     = nx
NP(2)     = ny
PERIOD(1) = .FALSE.
PERIOD(2) = .FALSE.
S_COL(2)  = .TRUE.
S_COL(1)  = .FALSE.
S_ROW(1)  = .TRUE.
S_ROW(2)  = .FALSE.
ORDER     = .TRUE.

CALL MPI_CART_CREATE(MPI_COMM_WORLD, 2, NP, PERIOD, ORDER, CART_GRP, IERR)
CALL MPI_CART_SUB(CART_GRP, S_ROW, row_comm, IERR)
CALL MPI_CART_SUB(CART_GRP, S_COL, col_comm, IERR)
CALL MPI_COMM_GROUP(row_comm, MEPGRP,        IERR)
CALL MPI_COMM_GROUP(col_comm, MEOGRP,        IERR)

CALL MPI_COMM_RANK(row_comm, row_rank, IERR)
CALL MPI_COMM_RANK(col_comm, col_rank, IERR)
CALL MPI_COMM_RANK(CART_GRP, RANK,     IERR)


!--------------------------------------------------------------------!
!Allocate-Generate matrix and right hand sides and time the procedure!
!--------------------------------------------------------------------!
p = RANK / ny
q = MOD(RANK, ny)

CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)
t1 = MPI_WTIME()

ALLOCATE(A(CHUNK1,CHUNK2))
A = 0.0D0

DO INDEX1=1, CHUNK2
  GINDEX1 = q*CHUNK2 + INDEX1   
  DO INDEX2=1, CHUNK1
     GINDEX2 = p*CHUNK1 + INDEX2
     IF (GINDEX1.EQ.GINDEX2) THEN
        A(INDEX2,INDEX1) = 1D0 + DBLE(GINDEX1)**theta
     ELSE
        A(INDEX2,INDEX1) = 1D0 / ( ABS(DBLE(GINDEX1) - DBLE(GINDEX2))**kappa )
     ENDIF
  ENDDO
ENDDO

!---------------------------!
!Initialize right-hand sides!
!---------------------------!
ALLOCATE(b(CHUNK2, rhs), x(CHUNK2, rhs))

CALL RANDOM_NUMBER(b)
DO I = 1, rhs
   DO J = 1, CHUNK2
      IF (b(J,I) .LE. 5D-1) THEN
         b(J, I) = -1D0
      ELSE
         b(J, I) = 1D0
      ENDIF
   ENDDO
ENDDO

CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)
t2 = MPI_WTIME()

IF (RANK == 0) THEN
   WRITE(*, *) " "
   WRITE(*, *) "==============================================================================="
   WRITE(*, *) "Matrix and right-hand sides were generated in", t2 - t1, "seconds"
   WRITE(*, *) "==============================================================================="
ENDIF


IF (RANK == 0) THEN 
   WRITE(*, *) " "
   WRITE(*, *) "==============================================================================="
   IF (1 == mkl_get_max_threads()) THEN 
      PRINT *, "Intel MKL utilizes one thread per MPI process" 
   ELSE 
      PRINT *, "Intel MKL utilizes", mkl_get_max_threads(), "threads per MPI process" 
   ENDIF
   WRITE(*, *) "==============================================================================="
   WRITE(*, *) " "
ENDIF

!==========================!
! Solve the linear systems !
!==========================!

!---------------------------------------------------------!
! choice 20  := only tests the MM product                 !
! choice 40  := executes all different BCG routines       !
!---------------------------------------------------------!


!===============================================================
! Distributed Conjugate-Gradient, each right-hand side is solved 
! on its own, but MV and inner products are aggregated
IF (choice == 1 .OR. choice == 40) THEN

   ALLOCATE(timings_cg(4))
   timings_cg = 0.0D0

   x = 0.0D0
   CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)
   t1 = MPI_WTIME()
   CALL distributed_cg(A, b(:,seedb1+1:rhs), x(:,seedb1+1:rhs), rhs-seedb1, tol2, CHUNK1, CHUNK2, N, nx, ny, maxit, cg_iters, timings_cg, RANK, CART_GRP, row_comm, col_comm)
   CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)
   t2 = MPI_WTIME()
   cg_matvecs    = cg_iters*(rhs-seedb1)
   MV_FLOPS      = cg_iters*(2*(rhs-seedb1)*(N**2.0D0)) 
   DOTAXPY_FLOPS = cg_iters*(10D0*N*((rhs-seedb1)))

   IF (RANK == 0) THEN
      WRITE(*, *) "==============================================================================="
      WRITE(*, *) "Number of right-hand sides solved", rhs-seedb1
      WRITE(*, *) "CG solved the linear system in", t2 - t1, "seconds"
      WRITE(*, *) "CG converged in", cg_iters, "iterations"
      WRITE(*, *) "Time spent on MATVECS = ", timings_cg(1)
      WRITE(*, *) "Time spent on DAXPYS + DOTS = ", timings_cg(2)
      WRITE(*, *) "Time spent on communication for the MATVECS = ", timings_cg(3)
      WRITE(*, *) "Time spent on communication for DOTS = ", timings_cg(4)
      WRITE(*, *) "Total MATVECS = ", cg_matvecs
      WRITE(*, *) "==============================================================================="
      WRITE(*, *) "GF/sec FOR THE MM OPERATION = ", DBLE(MV_FLOPS) / ((timings_cg(1)+timings_cg(3))*1.0D9)
      WRITE(*, *) "GF/sec FOR THE DOT+AXPY OPERATIONS = ", DBLE(DOTAXPY_FLOPS) / ((timings_cg(2)+timings_cg(4))*1.0D9)
      WRITE(*, *) "GF/sec TOTAL = ", DBLE(MV_FLOPS+DOTAXPY_FLOPS) / (SUM(timings_cg)*1.0D9)
      WRITE(*, *) "==============================================================================="
      WRITE(*, *) " "
      WRITE(*, *) " "
   ENDIF
   t2 = 0.0D0; t1 = 0.0D0

   DEALLOCATE(timings_cg)
ENDIF

!===========================================================
! Distributed Block Conjugate-Gradient, the "rhs" right-hand 
! sides are solved in batches of size "seedb2"
IF (choice == 2 .OR. choice == 40) THEN

   ALLOCATE(timings_bcg(5))
   timings_bcg = 0.0D0

   x = 0.0D0
   CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)
   t1 = MPI_WTIME()
   CALL distributed_bcg(A, b(:,seedb1+1:rhs), x(:,seedb1+1:rhs), rhs-seedb1, seedb2, tol2, CHUNK1, CHUNK2, N, nx, ny, maxit, bcg_iters, timings_bcg, RANK, CART_GRP, row_comm, col_comm)
   CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)
   t2 = MPI_WTIME()
   bcg_matvecs   = bcg_iters*seedb1
   MV_FLOPS      = bcg_iters*(2*(rhs-seedb1)*(N**2.0D0)) 
   DOTAXPY_FLOPS = bcg_iters*(10D0*N*((rhs-seedb1)**2.0D0)+3D0*N*(rhs-seedb1))

   IF (RANK == 0) THEN
      WRITE(*, *) "==============================================================================="
      WRITE(*, *) "Number of right-hand sides solved", rhs-seedb1
      WRITE(*, *) "Block CG solved the linear system in", t2 - t1, "seconds"
      WRITE(*, *) "Block CG converged in", bcg_iters, "iterations"
      WRITE(*, *) "Time spent on MATVECS = ", timings_bcg(1)
      WRITE(*, *) "Time spent on DAXPYS = ", timings_bcg(2)
      WRITE(*, *) "Time spent on DOTS = ", timings_bcg(3)
      WRITE(*, *) "Time spent on communication for the MATVECS = ", timings_bcg(4)
      WRITE(*, *) "Time spent on communication for DOTS = ", timings_bcg(5)
      WRITE(*, *) "Total MATVECS = ", bcg_matvecs
      WRITE(*, *) "==============================================================================="
      WRITE(*, *) "GF/sec FOR THE MM OPERATION = ", DBLE(MV_FLOPS) / ((timings_bcg(1)+timings_bcg(4))*1.0D9)
      WRITE(*, *) "GF/sec FOR THE DOT+AXPY OPERATIONS = ", DBLE(DOTAXPY_FLOPS) / ((timings_bcg(2)+timings_bcg(3)+timings_bcg(5))*1.0D9)
      WRITE(*, *) "GF/sec TOTAL = ", DBLE(MV_FLOPS+DOTAXPY_FLOPS) / (SUM(timings_bcg)*1.0D9)
      WRITE(*, *) "==============================================================================="
      WRITE(*, *) " "
      WRITE(*, *) " "
   ENDIF
   t2 = 0.0D0; t1 = 0.0D0

   DEALLOCATE(timings_bcg)
ENDIF

!================================================================
! Distributed Parallel-Projection Block Conjugate-Gradient --
! the first "seedb1" right-hand sides are solved by BCG and the
! generated Krylov subspace is used to accelerate convergence
! for the next batches of right-hand sides, each of size "seedb2"
IF (choice == 4 .OR. choice == 40) THEN

   ALLOCATE(timings_modinitbcg0(7))
   ALLOCATE(timings_bcg(5))
   timings_modinitbcg0 = 0.0D0
   timings_bcg         = 0.0D0

   ! Zero initial guesses
   x    = 0.0D0
   ! Maximum number of vectors allocated for each one of P and AP
   maxs = maxs*seedb1

   ! LIFO==1 --> reversed Galerkin projection
   ! LIFO==0 --> starndard Galerkin projection
   LIFO = 1D0

   ! "tau" controls how many direction blocks are simultaneously deflated,
   ! higher values will favor scalability but hinder the numerical part 
   ! default value: tau = 1D0
   tau = 1D0

   CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)
   t1 = MPI_WTIME()
   CALL distributed_modinitbcg(A, b, x, rhs, seedb1, tol1, CHUNK1, CHUNK2, N, nx, ny, maxit, maxs, modinitbcg_iters, LIFO, tau, timings_modinitbcg0, &
   RANK, CART_GRP, row_comm, col_comm)
   CALL distributed_bcg(A, b(:,seedb1+1:rhs), x(:,seedb1+1:rhs), rhs-seedb1, seedb2, tol2, CHUNK1, CHUNK2, N, nx, ny, maxit, &
   bcg_iters, timings_bcg, RANK, CART_GRP, row_comm, col_comm)
   CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)
   t2 = MPI_WTIME()
   modinitbcg_full_matvecs = modinitbcg_iters*seedb1 + bcg_iters*(rhs-seedb1)
   MV_FLOPS      = bcg_iters*(2*(rhs-seedb1)*(N**2.0D0)) + modinitbcg_iters*(2*seedb1*(N**2.0D0)) 
   DOTAXPY_FLOPS = bcg_iters*(10D0*N*((rhs-seedb1)**2.0D0)+3D0*N*(rhs-seedb1)) + modinitbcg_iters*(10D0*N*(seedb1**2.0D0)+3D0*N*seedb1)

   IF (RANK == 0) THEN
      WRITE(*, *) "==============================================================================="
      WRITE(*, *) "Galerkin projections: LIFO=", LIFO, "tau=", tau
      WRITE(*, *) "Size of the seed batch", seedb1
      WRITE(*, *) "Size of each following batch", seedb2
      WRITE(*, *) "PP-BCG solved the linear system in", t2 - t1, "seconds."
      WRITE(*, *) "-------------------------------------------------------------------------------"
      WRITE(*, *) "Seed converged in", modinitbcg_iters, "iterations."
      WRITE(*, *) "Total time spent on solving the seed system = ", timings_modinitbcg0(1) + timings_modinitbcg0(2) + timings_modinitbcg0(3) + timings_modinitbcg0(4) + timings_modinitbcg0(5)
      WRITE(*, *) "Time spent on MATVECS = ", timings_modinitbcg0(1)
      WRITE(*, *) "Time spent on DAXPYS = ", timings_modinitbcg0(2)
      WRITE(*, *) "Time spent on DOTS = ", timings_modinitbcg0(3)
      WRITE(*, *) "Time spent on communication for the MATVECS = ", timings_modinitbcg0(4)
      WRITE(*, *) "Time spent on communication for DOTS = ", timings_modinitbcg0(5)
      WRITE(*, *) "Total time spent on Galerkin Projections = ", timings_modinitbcg0(7) + timings_modinitbcg0(6)
      WRITE(*, *) "Time spent on communication for the seed phase = ", timings_modinitbcg0(7)
      WRITE(*, *) "Time spent on DAXPYS for the seed phase = ", timings_modinitbcg0(6)
      WRITE(*, *) "-------------------------------------------------------------------------------"
      WRITE(*, *) "BCG converged after", bcg_iters, "iterations."
      WRITE(*, *) "Total time spent on solving the nonseed system = ", timings_bcg(1) + timings_bcg(2) + timings_bcg(3) + timings_bcg(4) + timings_bcg(5)
      WRITE(*, *) "Time spent on MATVECS = ", timings_bcg(1)
      WRITE(*, *) "Time spent on DAXPYS = ", timings_bcg(2)
      WRITE(*, *) "Time spent on DOTS = ", timings_bcg(3)
      WRITE(*, *) "Time spent on communication for the MATVECS = ", timings_bcg(4)
      WRITE(*, *) "Time spent on communication for DOTS = ", timings_bcg(5)
      WRITE(*, *) "-------------------------------------------------------------------------------"
      WRITE(*, *) "Total time spent on MATVECS = ", timings_bcg(1) + timings_modinitbcg0(1)
      WRITE(*, *) "Total time spent on DAXPYS = ", timings_bcg(2) + timings_modinitbcg0(2) + timings_modinitbcg0(6)
      WRITE(*, *) "Total time spent on DOTS = ", timings_bcg(3) + timings_modinitbcg0(3)
      WRITE(*, *) "Total time spent on communication for MATVECS = ", timings_bcg(4) + timings_modinitbcg0(4)
      WRITE(*, *) "Total time spent on communication for DOTS = ", timings_bcg(5) + timings_modinitbcg0(5) + timings_modinitbcg0(7)
      WRITE(*, *) "Total time spent on communication = ", timings_bcg(4) + timings_modinitbcg0(4) + timings_bcg(5) + timings_modinitbcg0(5) + timings_modinitbcg0(7)
      WRITE(*, *) "-------------------------------------------------------------------------------"
      WRITE(*, *) "Total MATVECS = ", modinitbcg_full_matvecs
      WRITE(*, *) "==============================================================================="
      WRITE(*, *) "GF/sec FOR THE MM OPERATION = ", DBLE(MV_FLOPS) / ((timings_bcg(1)+timings_bcg(4)+timings_modinitbcg0(1)+timings_modinitbcg0(4))*1.0D9)
      WRITE(*, *) "GF/sec FOR THE DOT+AXPY OPERATIONS = ", DBLE(DOTAXPY_FLOPS)/((timings_bcg(2)+timings_bcg(3)+timings_bcg(5)+timings_modinitbcg0(2)+timings_modinitbcg0(3)+timings_modinitbcg0(5))*1.0D9)
      WRITE(*, *) "GF/sec TOTAL = ", DBLE(MV_FLOPS+DOTAXPY_FLOPS) / ( ( SUM(timings_bcg) + SUM(timings_modinitbcg0) )*1.0D9 )
      WRITE(*, *) "==============================================================================="
      WRITE(*, *) " "
   ENDIF
   t2 = 0.0D0; t1 = 0.0D0

   DEALLOCATE(timings_bcg, timings_modinitbcg0)
ENDIF

!===============================================
! This choice tests the MM routine that is going 
! to be used based on the 2-D grid topology
IF (choice == 20) THEN
   ALLOCATE(timings_mm(2))
   timings_mm = 0.0D0
   IF (nx .LT. ny) THEN
     CALL MM_2D_FAT(A, b, x, timings_mm, N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, CART_GRP, row_comm, col_comm)
   ENDIF
   IF (ny .LT. nx) THEN
     CALL MM_2D_SKINNY(A, b, x, timings_mm, N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, CART_GRP, row_comm, col_comm)
   ENDIF
   IF (ny .EQ. nx) THEN
     CALL MM_2D_SQUARE(A, b, x, timings_mm, N, CHUNK1, nx, rhs, RANK, CART_GRP, row_comm, col_comm)
   ENDIF

   IF (RANK == 0) THEN
     WRITE(*, *) "MM product: nx =, ny =, Comp Time =, Comm Time =", nx, ny, timings_mm(1), timings_mm(2)
   ENDIF
ENDIF


!--------------------------------------------------!
!Deallocate resources and terminate MPI environment!
!--------------------------------------------------!
DEALLOCATE(A, b, x)
CALL MPI_FINALIZE(IERR)

END PROGRAM mpi_mod_INITBCG_vI_2D




SUBROUTINE distributed_cg(A, b, x, rhs, tol, CHUNK1, CHUNK2, N, nx, ny, maxit, sumiter, timings, RANK, CART_GRP, row_comm, col_comm)

IMPLICIT NONE
INCLUDE 'mpif.h'

INTEGER,          INTENT(IN)                            :: rhs, CHUNK1, CHUNK2, N, nx, ny, maxit, RANK, CART_GRP, row_comm, col_comm
DOUBLE PRECISION, DIMENSION(CHUNK1, CHUNK2), INTENT(IN) :: A
DOUBLE PRECISION, DIMENSION(CHUNK2, rhs), INTENT(IN)    :: b
DOUBLE PRECISION, DIMENSION(CHUNK2, rhs), INTENT(INOUT) :: x
DOUBLE PRECISION, INTENT(IN)                            :: tol
INTEGER,          INTENT(OUT)                           :: sumiter
DOUBLE PRECISION, DIMENSION(4), INTENT(OUT)             :: timings

DOUBLE PRECISION, DIMENSION(rhs)                        :: ALPHA, BETA1, BETA, BETA11
DOUBLE PRECISION, DIMENSION(2*rhs)                      :: ALPHA1, ALPHA11
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE           :: p_cur, r_old, r_new, tmp, x_dummy, p_dummy
DOUBLE PRECISION, DIMENSION(:),   ALLOCATABLE           :: tmp_p_cur
DOUBLE PRECISION                                        :: BNM, RLNM, ddot, t1, t2
INTEGER                                                 :: IERR, I, J, row_rank, col_rank
INTEGER                                                 :: status(MPI_STATUS_SIZE) 
DOUBLE PRECISION, DIMENSION(2)                          :: times1, dummytime

!Allocations and initializations of local variables
sumiter = 0D0
timings = 0.0D0

ALLOCATE(p_cur(CHUNK2, rhs), r_old(CHUNK2, rhs), r_new(CHUNK2, rhs), tmp(CHUNK2, rhs))
ALLOCATE(tmp_p_cur(CHUNK2), p_dummy(CHUNK2, rhs), x_dummy(CHUNK2, rhs))

ALPHA1 = 0.0D0; ALPHA = 0.0D0; ALPHA11 = 0.0D0
BETA = 0.0D0; BETA1 = 0.0D0; BETA11 = 0.0D0
p_cur = 0.0D0; r_old = 0.0D0; r_new = 0.0D0 
tmp = 0.0D0; tmp_p_cur = 0.0D0
x_dummy = 0.0D0; p_dummy = 0.0D0

!Compute initial residual and initialize rest of the buffers
x_dummy = x
CALL MM_2D_WRAPPER(A, x_dummy, tmp, times1, N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, CART_GRP, row_comm, col_comm)
IF (RANK == 0) THEN; timings(1) = timings(1) + times1(1); timings(3) = timings(3) + times1(2); ENDIF

DEALLOCATE(x_dummy)

IF (RANK .LT. ny) THEN
   r_old(:, 1:rhs) = b(:, 1:rhs) - tmp(:, 1:rhs)
   CALL DCOPY(CHUNK2*rhs, r_old(1, 1), 1, r_new(1, 1), 1)
   CALL DCOPY(CHUNK2*rhs, r_old(1, 1), 1, p_cur(1, 1), 1)
ENDIF

CALL pddot(CHUNK2, b(1, 1), b(1, 1), BNM, dummytime, RANK, CART_GRP)
CALL pddot(CHUNK2, r_old(1, 1), r_old(1, 1), RLNM, dummytime, RANK, CART_GRP)
BNM = BNM / nx;

DO WHILE( (sqrt(RLNM)/sqrt(BNM)) >= tol .AND. sumiter < maxit )

    sumiter = sumiter + 1

    IF (RANK .EQ. 0) THEN
       !WRITE(*,*) "in iter", sumiter - 1, "rel. residual is", sqrt(RLNM) / sqrt(BNM)
    ENDIF

    !Perform distributed MM : A*P
    p_dummy = p_cur
    CALL MM_2D_WRAPPER(A, p_dummy, tmp, times1, N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, CART_GRP, row_comm, col_comm)    
    IF (RANK == 0) THEN; timings(1) = timings(1) + times1(1); timings(3) = timings(3) + times1(2); ENDIF

    !Find the \alpha, \beta scalars and update solution x and residual r
    CALL MPI_BARRIER(CART_GRP, IERR)
    IF (RANK == 0) THEN;  t1 = MPI_WTIME();  ENDIF
    ALPHA11 = 0.0D0; ALPHA1 = 0.0D0; ALPHA = 0.0D0
    IF (RANK .LT. ny) THEN
      DO I = 1, rhs
        ALPHA11(I) = DDOT(CHUNK2, r_old(1, I), 1, r_old(1, I), 1)
        ALPHA11(I+rhs) = DDOT(CHUNK2, p_cur(1, I), 1, tmp(1, I), 1)
      ENDDO
    ENDIF
    CALL MPI_BARRIER(CART_GRP, IERR)
    IF (RANK == 0) THEN;  t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

    CALL MPI_BARRIER(CART_GRP, IERR)
    IF (RANK == 0) THEN;  t1 = MPI_WTIME();  ENDIF
    CALL MPI_ALLREDUCE(ALPHA11, ALPHA1, 2*rhs, MPI_DOUBLE_PRECISION, MPI_SUM, CART_GRP, IERR)
    CALL MPI_BARRIER(CART_GRP, IERR)
    IF (RANK == 0) THEN;  t2 = MPI_WTIME(); timings(4) = timings(4) + (t2-t1); ENDIF

    CALL MPI_BARRIER(CART_GRP, IERR)
    IF (RANK == 0) THEN;  t1 = MPI_WTIME();  ENDIF
    BETA11 = 0.0D0; BETA = 0.0D0
    IF (RANK .LT. ny) THEN
      DO I = 1, rhs  
        ALPHA(I) = ALPHA1(I)/ALPHA1(I+rhs)
        CALL DAXPY(CHUNK2, ALPHA(I), p_cur(1, I), 1, x(1, I), 1)
        CALL DAXPY(CHUNK2, -ALPHA(I), tmp(1, I), 1, r_new(1, I), 1)
        BETA11(I) = DDOT(CHUNK2, r_new(1, I), 1, r_new(1, I), 1)
      ENDDO
    ENDIF
    CALL MPI_BARRIER(CART_GRP, IERR)
    IF (RANK == 0) THEN;  t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

    CALL MPI_BARRIER(CART_GRP, IERR)
    IF (RANK == 0) THEN;  t1 = MPI_WTIME();  ENDIF
    CALL MPI_ALLREDUCE(BETA11, BETA1, rhs, MPI_DOUBLE_PRECISION, MPI_SUM, CART_GRP, IERR)
    CALL MPI_BARRIER(CART_GRP, IERR)
    IF (RANK == 0) THEN;  t2 = MPI_WTIME(); timings(4) = timings(4) + (t2-t1); ENDIF

    !Update direction P and copy old residual to new residual
    CALL MPI_BARRIER(CART_GRP, IERR)
    IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
    IF (RANK .LT. ny) THEN
      DO I = 1, rhs
        BETA(I) = BETA1(I)/ALPHA1(I)
        CALL DCOPY(CHUNK2, p_cur(1, I), 1, tmp_p_cur, 1)
        CALL DCOPY(CHUNK2, r_new(1, I), 1, p_cur(1, I), 1)
        CALL DAXPY(CHUNK2, BETA(I), tmp_p_cur, 1, p_cur(1, I), 1)
        tmp_p_cur = 0.0D0
      ENDDO
      CALL DCOPY(CHUNK2*rhs, r_new(1, 1), 1, r_old(1, 1), 1)
    ENDIF
    CALL MPI_BARRIER(CART_GRP, IERR)
    IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

    IF (RANK == 0) THEN; RLNM = BETA1(1); ENDIF
    CALL MPI_BCAST(RLNM, 1, MPI_DOUBLE_PRECISION, 0, CART_GRP, IERR)
   
ENDDO

DEALLOCATE(p_cur, p_dummy, tmp)
DEALLOCATE(r_old, r_new, tmp_p_cur)

ENDSUBROUTINE distributed_cg

!--

SUBROUTINE distributed_bcg(A, b, x, rhs, seedb, tol, CHUNK1, CHUNK2, N, nx, ny, maxit, total_it, timings, RANK, CART_GRP, row_comm, col_comm)

IMPLICIT NONE
INCLUDE 'mpif.h'

!---------------------!
!Arguments declaration!
!---------------------!
INTEGER, INTENT(IN)                                     :: CHUNK1, CHUNK2, nx, ny, N, rhs, seedb, maxit, RANK, CART_GRP, row_comm, col_comm
DOUBLE PRECISION, DIMENSION(CHUNK1, CHUNK2), INTENT(IN) :: A
DOUBLE PRECISION, DIMENSION(CHUNK2, rhs), INTENT(IN)    :: b
DOUBLE PRECISION, DIMENSION(CHUNK2, rhs), INTENT(INOUT) :: x
DOUBLE PRECISION, INTENT(IN)                            :: tol
INTEGER, INTENT(OUT)                                    :: total_it
DOUBLE PRECISION, DIMENSION(5), INTENT(OUT)             :: timings

!Local variables
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE      :: r_old, r_new, p_cur2, x2, b2
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE      :: p_cur, out2, tmp2, x_dummy, p_dummy
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE      :: pTAp, rTr_old, out1, rTr_new, S, U, VT, pseud, t1
INTEGER                                            :: LDWORK
PARAMETER                                             (LDWORK=2000)
DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE        :: WORK
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE      :: tmp, IPIV, Z

DOUBLE PRECISION                                   :: RLNM, DDOT, BNM, t2, t3
INTEGER                                            :: I, IERR, INFO, K, J, local_it
INTEGER                                            :: status(MPI_STATUS_SIZE) 
DOUBLE PRECISION, DIMENSION(2)                     :: times1, times2

total_it = 0D0
local_it = 0D0

!------------------------------------------------!
!Allocation and initialization of local variables!
!------------------------------------------------!
ALLOCATE(p_cur(CHUNK2, seedb), out2(CHUNK2, seedb), p_cur2(CHUNK2, seedb), x_dummy(CHUNK2, rhs), p_dummy(CHUNK2, seedb))
ALLOCATE(pTAp(seedb, seedb), rTr_old(seedb, seedb), rTr_new(seedb, seedb))
ALLOCATE(out1(seedb, seedb), S(seedb, seedb), VT(seedb, seedb), U(seedb, seedb), pseud(seedb, seedb))
ALLOCATE(t1(seedb, seedb), Z(seedb, seedb), b2(CHUNK2, rhs))
ALLOCATE(WORK(LDWORK))
ALLOCATE(r_old(CHUNK2, rhs), r_new(CHUNK2, rhs), tmp2(CHUNK2, rhs), IPIV(CHUNK2, 1), tmp(CHUNK2, seedb))

x_dummy = 0.0D0; p_dummy = 0.0D0; b2 = 0.0D0
p_cur = 0.0D0; out2 = 0.0D0; p_cur2 = 0.0D0
pTAp = 0.0D0; rTr_old = 0.0D0; rTr_new = 0.0D0
out1 = 0.0D0; S = 0.0D0; U = 0.0D0; VT = 0.0D0; Z = 0.0D0
r_old = 0.0D0; r_new = 0.0D0; tmp = 0.0D0; IPIV = 0.0D0
tmp2 = 0.0D0

!Compute initial residual and initialize rest of the buffers
x_dummy = x
CALL MM_2D_WRAPPER(A, x_dummy, tmp2, times1, N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, CART_GRP, row_comm, col_comm)
IF (RANK == 0) THEN; timings(1) = timings(1) + times1(1); timings(4) = timings(4) + times1(2); ENDIF

DEALLOCATE(x_dummy)

b2 = b
IF (RANK .LT. ny) THEN
   r_old(1:CHUNK2, 1:rhs) = b2(1:CHUNK2, 1:rhs) - tmp2(1:CHUNK2, 1:rhs)
   r_new(1:CHUNK2, 1:rhs) = r_old(1:CHUNK2, 1:rhs)
ELSE
   b2 = 0.0D0
ENDIF

!Start BCG iteration independently on each block of size seedb
DO I = 1, rhs, seedb

  IF (RANK .LT. ny) THEN
     p_cur(1:CHUNK2, 1:seedb) = r_old(1:CHUNK2, I:I+seedb-1)
  ENDIF

  CALL pddot(CHUNK2, b2(1, I), b2(1, I), BNM, times2, RANK, CART_GRP)
  CALL pbdot(CHUNK2, r_old(1:CHUNK2, I:I+seedb-1), r_old(1:CHUNK2, I:I+seedb-1), seedb, rTr_old, times2, RANK, CART_GRP)
  IF (RANK == 0) THEN; RLNM = rTr_old(1,1); ENDIF
  CALL MPI_BCAST(RLNM, 1, MPI_DOUBLE_PRECISION, 0, CART_GRP, IERR)

  DO WHILE( (sqrt(RLNM)/sqrt(BNM)) >= tol .AND. local_it < maxit )

    !Increase iteration's counter and print current iteration's value and sq. root of relative residual
    total_it = total_it + 1; local_it = local_it + 1

    IF (RANK .EQ. 0) THEN
      !WRITE(*,*) "in iter", local_it - 1, "rel. residual is", sqrt(rTr_old(1, 1))/sqrt(BNM)
    ENDIF

    p_dummy = p_cur
    CALL MM_2D_WRAPPER(A, p_dummy, tmp, times1, N, CHUNK1, CHUNK2, nx, ny, seedb, RANK, CART_GRP, row_comm, col_comm)
    CALL MPI_BARRIER(CART_GRP, IERR)
    IF (RANK == 0) THEN; timings(1) = timings(1) + times1(1); timings(4) = timings(4) + times1(2); ENDIF

    CALL pbdot(CHUNK2, p_cur, tmp, seedb, pTAp, times2, RANK, CART_GRP)
    CALL MPI_BARRIER(CART_GRP, IERR)
    IF (RANK == 0) THEN; timings(3) = timings(3) + times2(1); timings(5) = timings(5) + times2(2); ENDIF

    IF (RANK .LT. ny) THEN
      IF (RANK == 0) THEN; t2 = MPI_WTIME(); ENDIF
      CALL DGESVD('A', 'A', seedb, seedb, pTAp, seedb, S, U, seedb, VT, seedb, WORK, LDWORK, INFO)

      Z = 0.0D0
      DO K = 1, seedb
         IF (S(K, 1) .NE. 0) THEN
            Z(K, K) = 1.0D0 / S(K, 1)
         ENDIF
      ENDDO

      CALL DGEMM('T', 'N', seedb, seedb, seedb, 1.0D0, VT, seedb, Z, seedb, 0.0D0, t1, seedb)
      CALL DGEMM('N', 'T', seedb, seedb, seedb, 1.0D0, t1, seedb, U, seedb, 0.0D0, pseud, seedb)
      CALL DGEMM('N', 'N', seedb, seedb, seedb, 1.0D0, pseud, seedb, rTr_old, seedb, 0.0D0, out1, seedb)

      !Update x, r
      CALL DGEMM('N', 'N', CHUNK2, seedb, seedb, 1.0D0, p_cur, CHUNK2, out1, seedb, 0.0D0, out2, CHUNK2)
      x(1:CHUNK2, I:I+seedb-1) = x(1:CHUNK2, I:I+seedb-1) + out2(1:CHUNK2, 1:seedb)
      out2 = 0.0D0
      CALL DGEMM('N', 'N', CHUNK2, seedb, seedb, 1.0D0, tmp, CHUNK2, out1, seedb, 0.0D0, out2, CHUNK2)
      r_new(1:CHUNK2, I:I+seedb-1) = r_old(1:CHUNK2, I:I+seedb-1) - out2(1:CHUNK2, 1:seedb)
      out2 = 0.0D0
      IF (RANK == 0) THEN; t3 = MPI_WTIME(); timings(2) = timings(2) + (t3-t2); ENDIF
    ENDIF

    !Compute pinv(r_old'r_old) and pinv(r_old'r_old)*(r_new'r_new)
    CALL pbdot(CHUNK2, r_new(1:CHUNK2,I:I+seedb-1), r_new(1:CHUNK2,I:I+seedb-1), seedb, rTr_new, times2, RANK, CART_GRP)
    CALL MPI_BARRIER(CART_GRP, IERR)
    IF (RANK == 0) THEN; timings(3) = timings(3) + times2(1); timings(5) = timings(5) + times2(2); ENDIF

    IF (RANK .LT. ny) THEN
      IF (RANK == 0) THEN; t2 = MPI_WTIME(); ENDIF
      CALL DGESVD('A', 'A', seedb, seedb, rTr_old, seedb, S, U, seedb, VT, seedb, WORK, LDWORK, INFO)

      Z = 0.0D0
      DO K = 1, seedb
         IF (S(K, 1) .NE. 0) THEN
            Z(K, K) = 1.0D0 / S(K, 1)
         ENDIF
      ENDDO

      CALL DGEMM('T', 'N', seedb, seedb, seedb, 1.0D0, VT, seedb, Z, seedb, 0.0D0, t1, seedb)
      CALL DGEMM('N', 'T', seedb, seedb, seedb, 1.0D0, t1, seedb, U, seedb, 0.0D0, pseud, seedb)
      CALL DGEMM('N', 'N', seedb, seedb, seedb, 1.0D0, pseud, seedb, rTr_new, seedb, 0.0D0, out1, seedb)
      rTr_old(1:seedb, 1:seedb) = rTr_new(1:seedb, 1:seedb)

      !Update P
      CALL DGEMM('N', 'N', CHUNK2, seedb, seedb, 1.0D0, p_cur, CHUNK2, out1, seedb, 0.0D0, out2, CHUNK2)
      p_cur(1:CHUNK2, 1:seedb) = r_new(1:CHUNK2, I:I+seedb-1) + out2(1:CHUNK2, 1:seedb)
      r_old(1:CHUNK2, I:I+seedb-1) = r_new(1:CHUNK2, I:I+seedb-1)
      IF (RANK == 0) THEN; t3 = MPI_WTIME(); timings(2) = timings(2) + (t3-t2); ENDIF
    ENDIF

    IF (RANK == 0) THEN; RLNM = rTr_old(1,1); ENDIF
    CALL MPI_BCAST(RLNM, 1, MPI_DOUBLE_PRECISION, 0, CART_GRP, IERR)

  ENDDO

  local_it = 0D0

ENDDO

DEALLOCATE(p_cur, out2, p_cur2)
DEALLOCATE(pTAp, rTr_old, rTr_new)
DEALLOCATE(out1, S, VT, U, pseud, Z)
DEALLOCATE(t1)
DEALLOCATE(WORK, p_dummy, b2)
DEALLOCATE(r_old, r_new, tmp, IPIV)

ENDSUBROUTINE distributed_bcg

!--

SUBROUTINE distributed_modinitbcg(A, b, x, rhs, seedb, tol, CHUNK1, CHUNK2, N, nx, ny, maxit, maxs, sumiter, LIFO, tau, timings, RANK, CART_GRP, row_comm, col_comm)

IMPLICIT NONE
INCLUDE 'mpif.h'

!---------------------!
!Arguments declaration!
!---------------------!
INTEGER, INTENT(IN)                                    :: CHUNK1, CHUNK2, N, rhs, seedb, maxit, maxs, LIFO, tau, CART_GRP, RANK, row_comm, col_comm, nx, ny
DOUBLE PRECISION, DIMENSION(CHUNK1,CHUNK2), INTENT(IN) :: A
DOUBLE PRECISION, DIMENSION(CHUNK2,rhs), INTENT(IN)    :: b
DOUBLE PRECISION, DIMENSION(CHUNK2,rhs), INTENT(INOUT) :: x
DOUBLE PRECISION, INTENT(IN)                           :: tol
INTEGER, INTENT(OUT)                                   :: sumiter
DOUBLE PRECISION, DIMENSION(7), INTENT(OUT)            :: timings

!Local variables
DOUBLE PRECISION, DIMENSION(:,:),ALLOCATABLE    :: r_old, r_new, x_dummy
DOUBLE PRECISION, DIMENSION(:,:),ALLOCATABLE    :: p_cur, out2, p_cur2, b2
DOUBLE PRECISION, DIMENSION(:,:),ALLOCATABLE    :: out2_mod, P_glob, AP_glob, pTAp_glob, rTr_glob, pTAp_TEMP
DOUBLE PRECISION, DIMENSION(:,:),ALLOCATABLE    :: pTAp, rTr_old, out1, rTr_new, S, VT, U, pseud, t1
DOUBLE PRECISION, DIMENSION(:,:),ALLOCATABLE    :: rTr_old_mod, H_mod
DOUBLE PRECISION, DIMENSION(:,:),ALLOCATABLE    :: ptap_store, Z
INTEGER                                         :: LDWORK, clock_start, clock_rate, clock_end
PARAMETER                                          (LDWORK=2000)
DOUBLE PRECISION, DIMENSION(:),ALLOCATABLE      :: WORK
DOUBLE PRECISION, DIMENSION(:,:),ALLOCATABLE    :: tmp, tmp2, x2

DOUBLE PRECISION                                :: RLNM, BNM, DDOT, t2, t3
INTEGER                                         :: I, IERR, INFO, K, J, re, val2
INTEGER                                         :: status(MPI_STATUS_SIZE) 
DOUBLE PRECISION, DIMENSION(2)                  :: times1, times2

! For the Galerkin projections' loop
INTEGER                                         :: total_steps, startt, stopp, stepp, col_rank

!DOUBLE PRECISION, DIMENSION(:,:),ALLOCATABLE    :: S_T, VT_T, U_T, Z_T, t1_T, pseud_T

CALL MPI_COMM_RANK(col_comm, col_rank, IERR)

val2 = int(maxs/seedb) ! seedb must divide maxs perfectly
sumiter = 0D0

!------------------------------------------------!
!Allocation and initialization of local variables!
!------------------------------------------------!
ALLOCATE(p_cur(CHUNK2, seedb), out2(CHUNK2, seedb), p_cur2(CHUNK2, seedb))
ALLOCATE(out2_mod(CHUNK2, rhs-seedb), P_glob(CHUNK2, maxs), AP_glob(CHUNK2, maxs), pTAp_glob(seedb, maxs), rTr_glob(seedb, maxs))
ALLOCATE(pTAp(seedb, seedb), rTr_old(seedb, seedb), rTr_new(seedb, seedb), pTAp_TEMP(seedb*tau, seedb*tau))
ALLOCATE(out1(seedb, seedb), S(seedb, seedb), VT(seedb, seedb), U(seedb, seedb), pseud(seedb, seedb))
ALLOCATE(t1(seedb, seedb))
ALLOCATE(WORK(LDWORK))
ALLOCATE(rTr_old_mod(tau*seedb, rhs-seedb), H_mod(tau*seedb, rhs-seedb), b2(CHUNK2, rhs))
ALLOCATE(pTAp_store(seedb, seedb), Z(seedb, seedb), tmp2(CHUNK2, rhs))
ALLOCATE(r_old(CHUNK2, rhs), r_new(CHUNK2, rhs), tmp(CHUNK2, seedb), x_dummy(CHUNK2, rhs))

!ALLOCATE(S_T(seedb*tau,seedb*tau),U_T(seedb*tau,seedb*tau),VT_T(seedb*tau,seedb*tau), Z_T(seedb*tau,seedb*tau), t1_T(seedb*tau,seedb*tau), pseud_T(seedb*tau,seedb*tau))

x_dummy = 0.0D0; b2 = 0.0D0
p_cur = 0.0D0; out2 = 0.0D0; p_cur2 = 0.0D0
out2_mod = 0.0D0
pTAp = 0.0D0; rTr_old = 0.0D0; rTr_new = 0.0D0; S = 0.0D0
out1 = 0.0D0
rTr_old_mod = 0.0D0; H_mod = 0.0D0
pTAp_store = 0.0D0; Z = 0.0D0
r_old = 0.0D0; r_new = 0.0D0; tmp = 0.0D0; tmp2 = 0.0D0
RLNM = 0.0D0; BNM = 0.0D0

!U_T = 0.0D0; S_T = 0.0D0; VT_T = 0.0D0; Z_T = 0.0D0; t1_T = 0.0D0; pseud_T = 0.0D0

!---------------------------------------------!
!Compute initial residual by performing r=b-Ax!
!---------------------------------------------!
x_dummy = x
CALL MM_2D_WRAPPER(A, x_dummy, tmp2, times1, N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, CART_GRP, row_comm, col_comm)
IF (RANK == 0) THEN; timings(1) = timings(1) + times1(1); timings(4) = timings(4) + times1(2); ENDIF

b2 = b
IF (RANK .LT. ny) THEN
   r_old(1:CHUNK2, 1:rhs) = b2(1:CHUNK2, 1:rhs) - tmp2(1:CHUNK2, 1:rhs)
   r_new(1:CHUNK2, 1:rhs) = r_old(1:CHUNK2, 1:rhs)
   p_cur(1:CHUNK2, 1:seedb) = r_old(1:CHUNK2, 1:seedb)
ELSE
   b2 = 0.0D0
ENDIF

DEALLOCATE(tmp2, x_dummy)

CALL pddot(CHUNK2, b2(:, 1), b2(:, 1), BNM, times2, RANK, CART_GRP)
CALL pbdot(CHUNK2, r_old(1:CHUNK2, 1:seedb), r_old(1:CHUNK2, 1:seedb), seedb, rTr_old, times2, RANK, CART_GRP)
IF (RANK == 0) THEN; RLNM = rTr_old(1,1); ENDIF
CALL MPI_BCAST(RLNM, 1, MPI_DOUBLE_PRECISION, 0, CART_GRP, IERR)

!-----------------!
!Main INITBCG loop!
!-----------------!
DO WHILE ( (sqrt(RLNM)/sqrt(BNM)) >=  tol .AND. sumiter < maxit )

    sumiter = sumiter + 1
    
    IF (RANK .LT. ny) THEN
       IF (sumiter <= val2) THEN; P_glob(1:CHUNK2, (sumiter-1)*seedb+1:sumiter*seedb) = p_cur; ENDIF
    ENDIF

    IF (RANK .EQ. 0) THEN
       !WRITE(*,*) "in iter", sumiter - 1, "rel. residual is", sqrt(RLNM)/sqrt(BNM)
    ENDIF

    !p_dummy = p_cur
    CALL MM_2D_WRAPPER(A, p_cur, tmp, times1, N, CHUNK1, CHUNK2, nx, ny, seedb, RANK, CART_GRP, row_comm, col_comm)
    IF (RANK == 0) THEN; timings(1) = timings(1) + times1(1); timings(4) = timings(4) + times1(2); ENDIF

    CALL pbdot(CHUNK2, p_cur, tmp, seedb, pTAp, times2, col_rank, col_comm)
    IF (RANK == 0) THEN; timings(3) = timings(3) + times2(1); timings(5) = timings(5) + times2(2); ENDIF

    IF (RANK .LT. ny) THEN
       IF (sumiter <= val2) THEN; AP_glob(1:CHUNK2, (sumiter-1)*seedb+1:sumiter*seedb) = tmp; ENDIF
    ENDIF

    IF (RANK .LT. ny) THEN
       IF (RANK == 0) THEN; t2 = MPI_WTIME(); ENDIF
       CALL DGESVD('A', 'A', seedb, seedb, pTAp, seedb, S, U, seedb, VT, seedb, WORK, LDWORK, INFO)
       Z = 0.0D0
       DO K = 1, seedb
          IF (S(K, 1) .NE. 0) THEN
            Z(K, K) = 1.0D0 / S(K, 1)
          ENDIF
       ENDDO

       CALL DGEMM('T', 'N', seedb, seedb, seedb, 1.0D0, VT, seedb, Z, seedb, 0.0D0, t1, seedb)
       CALL DGEMM('N', 'T', seedb, seedb, seedb, 1.0D0, t1, seedb, U, seedb, 0.0D0, pseud, seedb)
       ptap_store  = pseud
       IF (sumiter <= val2) THEN; pTAp_glob(1:seedb, (sumiter-1)*seedb+1:sumiter*seedb) = pseud; ENDIF
       CALL DGEMM('N', 'N', seedb, seedb, seedb, 1.0D0, pseud, seedb, rTr_old, seedb, 0.0D0, out1, seedb)

       !Update x, r
       CALL DGEMM('N', 'N', CHUNK2, seedb, seedb, 1.0D0, p_cur, CHUNK2, out1, seedb, 0.0D0, out2, CHUNK2)
       x(1:CHUNK2, 1:seedb) = x(1:CHUNK2, 1:seedb) + out2(1:CHUNK2, 1:seedb)
       out2 = 0.0D0
       CALL DGEMM('N', 'N', CHUNK2, seedb, seedb, 1.0D0, tmp, CHUNK2, out1, seedb, 0.0D0, out2, CHUNK2)
       r_new(1:CHUNK2, 1:seedb) = r_old(1:CHUNK2, 1:seedb) - out2(1:CHUNK2, 1:seedb)
       out2 = 0.0D0
       IF (RANK == 0) THEN; t3 = MPI_WTIME(); timings(2) = timings(2) + (t3-t2); ENDIF
    ENDIF

    !Compute pinv(r_old'r_old) and pinv(r_old'r_old)*(r_new'r_new)
    CALL pbdot(CHUNK2, r_new(1:CHUNK2, 1:seedb), r_new(1:CHUNK2, 1:seedb), seedb, rTr_new, times2, col_rank, col_comm)
    IF (RANK == 0) THEN; timings(3) = timings(3) + times2(1); timings(5) = timings(5) + times2(2); ENDIF

    IF (RANK .LT. ny) THEN
       IF (RANK == 0) THEN; t2 = MPI_WTIME(); ENDIF
       CALL DGESVD('A', 'A', seedb, seedb, rTr_old, seedb, S, U, seedb, VT, seedb, WORK, LDWORK, INFO)
 
       Z = 0.0D0
       DO K = 1, seedb
          IF (S(K, 1) .NE. 0) THEN
             Z(K, K) = 1.0D0 / S(K, 1)
          ENDIF
       ENDDO

       CALL DGEMM('T', 'N', seedb, seedb, seedb, 1.0D0, VT, seedb, Z, seedb, 0.0D0, t1, seedb)
       CALL DGEMM('N', 'T', seedb, seedb, seedb, 1.0D0, t1, seedb, U, seedb, 0.0D0, pseud, seedb)
       CALL DGEMM('N', 'N', seedb, seedb, seedb, 1.0D0, pseud, seedb, rTr_new, seedb, 0.0D0, out1, seedb)
       rTr_old(1:seedb, 1:seedb) = rTr_new(1:seedb, 1:seedb)
       IF (sumiter <= val2) THEN; rTr_glob(1:seedb, (sumiter-1)*seedb+1:sumiter*seedb) = out1; ENDIF

       !Update P
       CALL DGEMM('N', 'N', CHUNK2, seedb, seedb, 1.0D0, p_cur, CHUNK2, out1, seedb, 0.0D0, out2, CHUNK2)
       p_cur(1:CHUNK2, 1:seedb) = r_new(1:CHUNK2, 1:seedb) + out2(1:CHUNK2, 1:seedb)
       r_old(1:CHUNK2, 1:seedb) = r_new(1:CHUNK2, 1:seedb)
       IF (RANK == 0) THEN; t3 = MPI_WTIME(); timings(2) = timings(2) + (t3-t2); ENDIF
    ENDIF

    
    IF (RANK == 0) THEN; RLNM = rTr_old(1,1); ENDIF
    CALL MPI_BCAST(RLNM, 1, MPI_DOUBLE_PRECISION, 0, CART_GRP, IERR)

ENDDO

!--------------------!
!Galerkin projections!
!--------------------!
! total_steps denotes how many projection steps will be performed,
! each step deflates tau direction blocks
IF (sumiter .GT. val2) THEN 
  sumiter=val2
ENDIF
total_steps = sumiter/tau

IF (LIFO .EQ. 0D0) THEN
   startt = 1D0; stopp = total_steps; stepp = tau
ELSE
   startt = total_steps; stopp = 1D0; stepp = -tau
ENDIF


DO I = startt, stopp, stepp

  DO J = 1, tau
    pTAp_TEMP((J-1)*seedb+1:J*seedb, (J-1)*seedb+1:J*seedb) = pTAp_glob(:, (I-1)*tau*seedb+(J-1)*seedb+1:(I-1)*tau*seedb+J*seedb)
  ENDDO

  CALL pbdot2(CHUNK2, P_glob(1:CHUNK2, (I-1)*tau*seedb+1:I*tau*seedb), r_old(1:CHUNK2, seedb+1:rhs), seedb*tau, rhs-seedb, rTr_old_mod, times2, col_rank, col_comm)
  IF (RANK == 0) THEN; timings(6) = timings(6) + times2(2); timings(7) = timings(7) + times2(1); ENDIF

  IF (RANK .LT. ny) THEN
     IF (RANK == 0) THEN; t2 = MPI_WTIME(); ENDIF
     CALL DGEMM("N","N", seedb*tau, (rhs-seedb), seedb*tau, 1.0D0, pTAp_TEMP, tau*seedb, rTr_old_mod, seedb*tau, 0.0D0, H_mod, seedb*tau)
     CALL DGEMM("N", "N", CHUNK2, (rhs-seedb), seedb*tau, 1.0D0, P_glob(1:CHUNK2, (I-1)*tau*seedb+1:I*tau*seedb), CHUNK2, H_mod, seedb*tau, 0.0D0, out2_mod, CHUNK2)
     x(:, seedb+1:rhs) = x(:, seedb+1:rhs) + out2_mod
     CALL DGEMM("N", "N", CHUNK2, (rhs-seedb), seedb*tau, 1.0D0, AP_glob(1:CHUNK2, (I-1)*tau*seedb+1:I*tau*seedb), CHUNK2, H_mod, seedb*tau, 0.0D0, out2_mod, CHUNK2)
     r_new(:, seedb+1:rhs) = r_old(:, seedb+1:rhs) - out2_mod
     r_old(:, seedb+1:rhs) = r_new(:, seedb+1:rhs)
     IF (RANK == 0) THEN; t3 = MPI_WTIME(); timings(6) = timings(6) + (t3-t2); ENDIF
  ENDIF

ENDDO

!------------------!
!De-allocate memory!
!------------------!
DEALLOCATE(p_cur, out2)
DEALLOCATE(out2_mod, p_cur2, pTAp_glob, P_glob, AP_glob, rTr_glob)
DEALLOCATE(pTAp, rTr_old, rTr_new, pTAp_TEMP)
DEALLOCATE(out1, S, VT, U, pseud, Z)
DEALLOCATE(t1)
DEALLOCATE(WORK)
DEALLOCATE(rTr_old_mod, H_mod)
DEALLOCATE(pTAp_store)
DEALLOCATE(r_old, r_new, tmp)

END SUBROUTINE distributed_modinitbcg
